/*-------------------------------------------------------------------
Name......... game.c
Author....... Mike Ballinger & Darren Williams
Date......... 15/09/03
Description.. Contains boot loader code and calls security function
-------------------------------------------------------------------*/

/*-------------------------------------------------------------------
DECLARATION FORMATS
Type defs........... UPPER CASE
#defines............ UPPER CASE
Global Functions.... GlobalFunction();
Static Functions.... staticfunction();
Global Variables.... GlobalVariable;
Static Variables.... staticvariable;
Local Variables..... localvariable;
Structure Elements.. Structure.element;
-------------------------------------------------------------------*/

/****   Include Files */

#include    "hardware.h"
#include    "peripherals.h"
#include    "interface.h"
#include    "config.h"
#include    "fpga.h"

/****   Local Definitions */
/****   Global Variables */
/****   Local Variables */
/****   Global Constants */

#define BYTES_IN_SECTOR 512 /* Number of bytes in a file sector */
#define WORDS_IN_SECTOR (BYTES_IN_SECTOR / 2)

typedef struct bootloader
{
    FILE_DESCRIPTOR file;               /* File pointer */
    LONG    firstAddress;               /* First address in S19 file */
    ULONG   s19FilePos;                 /* Current file offset */
    BYTE    fileBuffer[BYTES_IN_SECTOR];   /* Buffer containing 1 sector
                                           of file data */
    LONG    fileBufferPos;              /* Position in file buffer */
} BOOTLOADER;

/* Assembly functions defined in entry.s */
extern  void    SetStackPointer(LONG);
extern  void    JumpToLocation(LONG);

/****   Identity */

const LONG  MachineId=0x00000002;

/****   Local Constants */
/************************************************************************
*****   Functions
*************************************************************************/



/* Macros to speed up processing of S19 File */

/*-------------------------------------------------------------------
Name......... Macro_CharConvert()
Author....... Mike Ballinger (converted to Macro by Darren Williams)
Date......... 17/02/04
Description.. MACRO:
............. Converts a Nibble (0-A) to an actual value
Parameters... nib - decoded number returned
Return....... none
History...... Based on NibblesToNum() 05/01/04 - MJB
............. 18/02/04 - Simplified and Converted to Macro form - DW
-------------------------------------------------------------------*/

#define MacroCharConvert(nib)                                       \
                                                                    \
                if(( nib >= '0' ) && ( nib<='9' ))                  \
                {                                                   \
                    nib = nib - '0';                                \
                }                                                   \
                else if((nib >= 'A') && (nib <= 'F'))               \
                {                                                   \
                    nib = nib - 'A' + 0x0a;                         \
                }                                                   \
                else if (( nib >= 'a' ) && ( nib <= 'f' ))          \
                {                                                   \
                    nib = nib - 'a' + 0x0a;                         \
                }

/*-------------------------------------------------------------------
Name......... Macro_GetS19FileByte(nib)
Author....... Mike Ballinger (converted to Macro by Darren Williams)
Date......... 17/02/04
Description.. MACRO:
............. Obtains a nibble from the specified S19 file.
............. Reads a new sector from file if required.
Parameters... nib - the nibble from the S19 file returned
Return....... none
History...... Based on GetS19FileByte() 05/01/04 - MJB
............. 18/02/04 - Simplified and converted to Macro form - DW
-------------------------------------------------------------------*/

#define MacroGetS19FileByte(nib)                                    \
                                                                    \
                position = bootLoader->fileBufferPos;               \
                                                                    \
                nib = (ULONG)bootLoader->fileBuffer[position++];    \
                                                                    \
                if(position == BYTES_IN_SECTOR)                        \
                {                                                   \
                    if(LoadS19Sector(bootLoader)!=0)                \
                    {                                               \
                        return(1);                                  \
                    }                                               \
                }                                                   \
                else                                                \
                {                                                   \
                    bootLoader->fileBufferPos = position;           \
                }                                                   \
                                                                    \
                bootLoader->s19FilePos ++;

/*-------------------------------------------------------------------
Name......... MacroS19ToDataByte()
Author....... Darren Williams
Date......... 17/02/04
Description.. MACRO:
............. Obtains 2 bytes from S19 file and converts to a single
............. data byte
Parameters... result - LONG value containing 1 data byte
............. nib    - nibble value
Return....... none
History...... Based on GetS19FileByte() 05/01/04 - MJB
............. 18/02/04 - Macro provides equivalent functionality to:
.............            result = NibblesToNum(bootLoader, 2); -DW
-------------------------------------------------------------------*/

#define MacroS19ToDataByte(result, nib)                             \
                                                                    \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = nib;                                       \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;

/*-------------------------------------------------------------------
Name......... MacroS19To4DataBytes()
Author....... Darren Williams
Date......... 17/02/04
Description.. MACRO:
............. Obtains 8 bytes from S19 file and converts to four
............. data bytes
Parameters... result - LONG value containing 4 data bytes
............. nib    - nibble value
Return....... none
History...... Based on GetS19FileByte() 05/01/04 - MJB
............. 18/02/04 - Macro provides equivalent functionality to:
.............            result = NibblesToNum(bootLoader, 8); - DW
-------------------------------------------------------------------*/

#define MacroS19To4DataBytes(result, nib)                           \
                                                                    \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = nib;                                       \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;                                      \
                MacroGetS19FileByte(nib)                            \
                MacroCharConvert(nib)                               \
                result = result << 4;                               \
                result += nib;


/*-------------------------------------------------------------------
Name......... _putchar()
Author....... Marc Adams
Date......... 15/09/03
Description.. Fuction stub to satisfy GNU compiler
Parameters... None
Return....... None
History...... 15/09/03 - Code tidy up  - MSA
-------------------------------------------------------------------*/

void    _putchar()
{

}



/*-------------------------------------------------------------------
Name......... EndOfS19File()
Author....... Mike Ballinger
Date......... 05/01/04
Description.. Checks to see if we have reached the end of the S19
............. file.
Parameters... bootLoader - Pointer to structure of type BOOTLOADER
Return....... 1 if end of S19 file, 0 otherwise
History...... 05/01/04 - Initial version - MJB
-------------------------------------------------------------------*/

BYTE EndOfS19File(BOOTLOADER *bootLoader)
{
    if(bootLoader->s19FilePos >= bootLoader->file.size)
    {
        return (1);
    }

    return( 0 );
}


/*-------------------------------------------------------------------
Name......... LoadS19Sector()
Author....... Mike Ballinger
Date......... 05/01/04
Description.. Loads a single sector from the S19 file
Parameters... bootLoader - Pointer to structure of type BOOTLOADER
Return....... 1 on read error, 0 on success
History...... 05/01/04 - Initial version - MJB
-------------------------------------------------------------------*/

BYTE LoadS19Sector(BOOTLOADER *bootLoader)
{
    UWORD   *infowordmemptr;

    (ULONG)infowordmemptr = (ULONG)&bootLoader->fileBuffer[0];

    /* Read S19 file sector */
    if(FileGetSectors(NULL, &bootLoader->file, infowordmemptr, 1) != 1)
    {
        return(1);
    }

    bootLoader->fileBufferPos = 0;

    return(0);
}


/*-------------------------------------------------------------------
Name......... AnalyseS19File()
Author....... Mike Ballinger
Date......... 18/02/04
Description.. Analyse the S19 file, 1 byte at a time, and store each
............. BYTE in its correct memory location.
............. Will not check if destination address is valid (DW)
Parameters... bootLoader - Pointer to structure of type BOOTLOADER
Return....... 0 on success, 1 otherwise
History...... 05/01/04 - Initial version - MJB
............. 18/02/04 - Code optimisation for speed.
.............            Function now uses register variables to
.............            increase speed and Macros are used in
.............            place of selected function calls - DW
-------------------------------------------------------------------*/

BYTE AnalyseS19File(BOOTLOADER *bootLoader)
{
    /* Make variables registered to increase speed */
    /* Most critical variable declared at top:- */
    register LONG nibble;
    register LONG position;
    register LONG calcChecksum;
    register LONG dataByte;
    register LONG address;
    register LONG fileChecksum;
    register LONG dataRemaining;
    register UBYTE *destinationAddress;
    register LONG dataLength;
    register LONG firstAddressStored;

    firstAddressStored = 0;

    while(1)
    {
        /* Exit if we have reached end of file */
        if(EndOfS19File(bootLoader))
        {
            break;
        }

        /* Check this line begins with a valid S-Record */
        /* Get sRecordFirstByte */
        nibble = 0;
        MacroGetS19FileByte(nibble)

        if(nibble == 'S')
        {
            /* Get sRecordType */
            nibble = 0;
            MacroGetS19FileByte(nibble)

            switch(nibble)
            {
			case '3' :
				/*
				Obtain line length. This is the first byte
				stored in checksum
				*/
				calcChecksum=0;
				MacroS19ToDataByte(calcChecksum, nibble)

				/* Subtract 5 from line length to get data
				length (4 for address, 1 for checksum)
				*/
				dataLength = calcChecksum - 5;

				MacroS19To4DataBytes(address, nibble)
				destinationAddress = (BYTE *)address;

				if(firstAddressStored == 0)
				{
					bootLoader->firstAddress = address;
					firstAddressStored = 1;
				}

				calcChecksum += (address & 0xff);
				calcChecksum += ((address >> 8) & 0xff);
				calcChecksum += ((address >> 16) & 0xff);
				calcChecksum += ((address >> 24) & 0xff);

				dataRemaining = dataLength;

				while(dataRemaining)
				{

					dataByte=0;
					MacroS19ToDataByte(dataByte, nibble)
					*destinationAddress++ = dataByte;
					dataRemaining --;
				}

				/* Read back data for checksum comparison */
				dataRemaining = dataLength;
				destinationAddress = (BYTE *)address;
				while(dataRemaining)
				{
					calcChecksum += *destinationAddress++;
					dataRemaining --;
				}

				calcChecksum = 255 - (calcChecksum % 256);

				fileChecksum = 0;
				MacroS19ToDataByte(fileChecksum, nibble)

				/* Check for valid checksum. If not, return */
				if(calcChecksum != fileChecksum)
				{
					return(1);
				}
				break;

			case '0' :
			case '7' :
				/*
				Obtain line length. This is the first
				byte stored in checksum
				*/
				calcChecksum = 0;
				MacroS19ToDataByte(calcChecksum, nibble)

				/*
				Subtract 3 from line length to get data length
				(2 for address, 1 for checksum
				*/
				dataLength = calcChecksum - 1;

				while(dataLength)
				{
					dataByte = 0;
					MacroS19ToDataByte(dataByte, nibble)
					calcChecksum += dataByte;
					dataLength --;
				}

				calcChecksum = 255 - (calcChecksum % 256);

				fileChecksum = 0;
				MacroS19ToDataByte(fileChecksum, nibble)

				/* Check for valid checksum. If not, return */
				if(calcChecksum != fileChecksum)
				{
					return(1);
				}
				break;

			default :
				/* Unknown S-Record type */
				return(1);
				break;
            }

            /* Must remove newline */
            MacroGetS19FileByte(nibble)     /* 13 */
            MacroGetS19FileByte(nibble)     /* 10 */
        }
        else return(1);
    }

    return(0);
}

/*-------------------------------------------------------------------
Name......... ExecuteS19File()
Author....... Mike Ballinger
Date......... 18/02/04
Description.. Loads the required S19 file and sets function
............. arguments stackPointer and codePointer so that the
............. calling function can execute S19 file loaded by using
............. the functions SetStackPointer() and JumpToLocation()
Parameters... filename     - e.g. "dram_sec.s19"
............. stackPointer - value is read from S19 file loaded
............. codePointer  - value is read from S19 file loaded
Return....... 1 on error,
			  0 on sucess
History...... 17/12/03 - Initial version - MJB
............. 18/02/04 - Changed so doesn't execute S19 file but
.............            loads and returns values to execute if
.............            execution of code in DRAM is required - DW
............. 10/08/05 - Fixed warning about missing return at the bottom
						 of the function
-------------------------------------------------------------------*/

BYTE ExecuteS19File(BYTE *filename, LONG *stackPointer, LONG *codePointer)
{
    BOOTLOADER bootLoader;
    LONG *adrPtr;

    /* Open file */
    if(FileOpen( NULL, filename, &bootLoader.file, READ_ONLY) == NULL)
    {
        return(1);
    }

    /* Set current S19 file offset to 0 */
    bootLoader.s19FilePos = 0;

    /* Load initial S19 data */
    if(LoadS19Sector(&bootLoader) != 0)
    {
        return(1);
    }

    /* Analyse required S19 file, and return result */
    if(AnalyseS19File(&bootLoader) != 0)
    {
        return(1);
    }

    /* Close File */
    FileClose(&bootLoader.file);

    /*
    Jump to correct memory location. The STACK pointer is located in
    the first location specified in the S19 file, and the memory
    location to jump to is the next LONG (e.g. STACK PTR + 4)
    */

    adrPtr = (LONG *)bootLoader.firstAddress;
    *stackPointer = *adrPtr;
    *codePointer = *(adrPtr + 1);

	return(0);
}

/*-------------------------------------------------------------------
Name......... GraphicsCardDetected()
Author....... Mike Ballinger
Date......... 06/01/04
Description.. Checks to see if a graphics card is fitted
Parameters... None
Return....... 1 if card fitted, 0 if not
History...... 06/01/04 - Initial version - MJB
-------------------------------------------------------------------*/

BYTE GraphicsCardDetected()
{
    UWORD mmr;

    /*
    Set Chip Select Control Register 3 (CSCR3) to a specific
    wait state. This prevents the Coldfire hanging when reading
    Cremson_MMR if a graphics card is not plugged in.
    Bits set:
      WS0:WS3 (bits 10:13) Wait state, set to 10
      AA      (bit 8)      Auto Acknowledge enabled
      ASET    (bit 4)      Address Setup Enable
      WR      (bit 1)      Write Enable
      RD      (bit 0)      Read Enable
    */
    REGWV(MBAR+0x92) = 0x2913;

    mmr = REGWV(Cremson_MMR);   /* sometimes incorrect on first read */
    mmr = REGWV(Cremson_MMR);

    REGWV(MBAR+0x92) = 0x13;

    if(mmr == 0xebb9)
    {
        return(1);
    }

    return(0);
}

/*-------------------------------------------------------------------
Name......... Game()
Author....... Mike Ballinger
Date......... 15/09/03
Description.. Boot loader
Parameters... None
Return....... None
History...... 17/012/03 - Initial version - MJB
-------------------------------------------------------------------*/

void Game()
{

    BYTE retvalue=0;

    LONG stackPointer;
    LONG codePointer;

    /****   Security */
    /* Secure Address Range 1 start and end address */
    ULONG secure1start  = Sec1Start;
    ULONG secure1end    = Sec1End;
    /* Secure Address Range 2 start and end address */
    ULONG secure2start  = Sec2Start;
    ULONG secure2end    = Sec2End;

    /* Load S19 File from Compact Flash to DRAM */

    /* Must begin by initialising Compact Flash */
    if(ReadDriveID(&ATADevice) == 0)
    {
        /* Mount Compact Flash */
        if(MountFAT32( &ATADevice) == 0)
        {
            /*
            Load S19 file from CFlash to DRAM.
            Ensure correct file is specified.
            This file can either be secure or unsecure:
            - If the file is secured then ensure that secure address
              ranges are set correctly in config.c.
            - If file is not secured, ensure secure address ranges in
              config.c are all set to 0
            */

            if(GraphicsCardDetected( ))
            {
                /* Load video S19 software */
                retvalue = ExecuteS19File("vdemo.s19",
                                          &stackPointer, &codePointer);
            }
            else
            {
                /* Load non-video S19 software */
                retvalue = ExecuteS19File("demo.s19",
                                          &stackPointer, &codePointer);
            }
        }
    }

    if (retvalue == 1)
    {
        /*
        If we reach this point a problem has occurred.
        This will be either:
        1) There was no compact flash card fitted.
        2) The compact flash card is formatted incorrectly and could not
           be mounted.
        3) The required file could not be found on the flash card
        4) There was a problem during file load
           (for example, an invalid checksum)
        */

        while(1);
    }

    /* Run security function */
    //UnlockSecurity(secure1start, secure1end, secure2start, secure2end);

    /*
    Security now unlocked, to run from DRAM:-
    Jump to correct memory location set by function LoadS19File().
    The STACK pointer is located in the first location specified in
    the S19 file, and the memory location to jump to is the next
    LONG (e.g. STACK PTR + 4)
    */
    SetStackPointer(stackPointer);
    JumpToLocation(codePointer);

    /*
    If we reach this point a problem has occurred with trying to run
    from DRAM
    */

    /* Just do nothing. */
    while(1);
}

/*******************************************************************/